from collections import OrderedDict
from datetime import datetime
from decimal import Decimal

from pydantic_core._pydantic_core import Url
from yaml import Node, ScalarNode, YAMLError, dump, load
from yaml.representer import Representer as BaseRepresenter
from yaml.representer import SafeRepresenter

try:
    from yaml import CDumper as Dumper
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Dumper, Loader  # type: ignore


__all__ = ["yaml_loads", "yaml_dumps"]

__author__ = "Md Nazrul Islam"
__email__ = "email2nazrul@gmail.com"


class Representer(SafeRepresenter):
    """ """

    def represent_decimal(self, data: Decimal) -> Node:
        """ """
        coerced = float(data)
        return self.represent_float(coerced)

    def represent_datetime(self, data: datetime) -> ScalarNode:
        """Issue #96, we are using T separator for ISO format"""
        value = data.isoformat(sep="T")
        return self.represent_scalar("tag:yaml.org,2002:timestamp", value)

    def represent_url(self, data: Url) -> Node:
        """!!python/object/new:pydantic_core._pydantic_core.Url"""
        coerced = str(data)
        return self.represent_str(coerced)


BaseRepresenter.add_representer(
    Decimal, Representer.represent_decimal  # type: ignore[arg-type]
)
# important! we don't want explicit tag of OrderedDict
# (tag:yaml.org,2002:python/object/apply:collections.OrderedDict)
BaseRepresenter.add_representer(OrderedDict, SafeRepresenter.represent_dict)
BaseRepresenter.add_representer(
    datetime, Representer.represent_datetime  # type: ignore[arg-type]
)
BaseRepresenter.add_representer(
    Url, Representer.represent_url  # type: ignore[arg-type]
)


def yaml_loads(stream, loader=None):
    """ """
    loader = loader or Loader
    try:
        return load(stream, Loader=loader)
    except YAMLError as exc:
        # ensure Pydantic compatible error handling
        raise ValueError(f"YAMLError: {exc}")


def yaml_dumps(
    data,
    *,
    stream=None,
    indent=None,
    width=None,
    line_break=None,
    sort_keys=False,
    encoding="utf-8",
    return_bytes=True,
):
    """ """
    try:
        res = dump(
            data,
            stream=stream,
            Dumper=Dumper,
            indent=indent,
            width=width,
            line_break=line_break,
            sort_keys=sort_keys,
            encoding=encoding,
        )
    except YAMLError as exc:
        # ensure Pydantic compatible error handling
        raise ValueError(f"YAMLError: {exc}")

    if stream is None and return_bytes is False:
        res = res.decode()

    return res



import typing
from functools import lru_cache
from typing import get_args, get_origin

from pydantic._internal._validators import import_string
from pydantic.fields import FieldInfo
from pydantic.types import Base64Encoder, EncodedBytes

from .constraints import FHIR_PRIMITIVES_MAPS, FHIR_TYPES_MAPS, PY_PRIMITIVES

__author__ = "Md Nazrul Islam"
__email__ = "email2nazrul@gmail.com"


def _is_primitive_type(annotation: typing.Any) -> typing.Union[bool, None]:
    """ """
    origin = get_origin(annotation)
    if origin is None:
        if annotation in FHIR_PRIMITIVES_MAPS:
            return True

    args = get_args(annotation)
    for arg in args:
        if _is_primitive_type(arg) is not None:
            # check the number of args
            return True
    return None


@lru_cache(maxsize=1024, typed=True)
def is_primitive_type(field: FieldInfo) -> bool:
    """ """
    return _is_primitive_type(field.annotation) is not None


def _is_list_type(annotation: typing.Any) -> typing.Union[bool, None]:
    """ """
    origin = get_origin(annotation)
    if origin is None:
        return annotation is list
    if origin is list:
        return True
    args = get_args(annotation)
    for arg in args:
        val = _is_list_type(arg)
        if val is not None:
            return val
    return None


@lru_cache(maxsize=1024, typed=True)
def is_list_type(field: FieldInfo) -> bool:
    """ """
    return _is_list_type(field.annotation) is True


def _get_fhir_type(annotation: typing.Any, field: FieldInfo):
    """ """

    def _get(anno):
        if anno in FHIR_PRIMITIVES_MAPS:
            if anno in PY_PRIMITIVES:
                for type_ in field.metadata:
                    if type_.__class__ in FHIR_PRIMITIVES_MAPS:
                        return type_.__class__
            return anno
        try:
            if anno.__name__ in FHIR_TYPES_MAPS:
                return anno.__name__
        except AttributeError:
            pass
        return None

    val = _get(annotation)
    if val is not None:
        return val

    origin = get_origin(annotation)
    if origin is None:
        return _get(annotation)

    args = get_args(annotation)
    for arg in args:
        val = _get_fhir_type(arg, field)
        if val is not None:
            # check the number of args
            return val
    return None


def get_fhir_type_name(field: FieldInfo):
    """ """
    fhir_type = _get_fhir_type(field.annotation, field)
    if fhir_type is None:
        raise ValueError
    if fhir_type in FHIR_PRIMITIVES_MAPS:
        return FHIR_PRIMITIVES_MAPS[fhir_type]
    if fhir_type in FHIR_TYPES_MAPS:
        return import_string(FHIR_TYPES_MAPS[fhir_type]).get_resource_type()
    raise ValueError


@lru_cache(maxsize=1024, typed=True)
def get_base64_encoder(field_info: FieldInfo) -> typing.Any:
    """ """
    for enc in field_info.metadata:
        # handle base64 output
        if isinstance(enc, EncodedBytes):
            return enc
    if "Base64Binary" in str(field_info.annotation):
        return Base64Encoder
    return None


__all__ = [
    "is_primitive_type",
    "get_fhir_type_name",
    "is_list_type",
    "get_base64_encoder",
]
